import asyncio
from base64 import urlsafe_b64decode, urlsafe_b64encode
from datetime import date, datetime
from json import JSONEncoder, dumps, loads
from typing import overload
from typing_extensions import override
from iso8601 import parse_date
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from aiofauna.objects import FaunaTime, Native, Query, Ref, SetRef
from aiofauna.query import Expr
from aiohttp.web import Request, Response


def parse_json(json_string):
    """

    Parses a JSON string into python values.

    Also parses :any:`Ref`, :any:`SetRef`, :any:`FaunaTime`, and :class:`date`.
    """

    return loads(json_string, object_hook=_parse_json_hook)


def parse_json_or_none(json_string):
    try:
        return parse_json(json_string)

    except ValueError:
        pass


def _parse_json_hook(dct):
    # pylint: disable=too-many-return-statements
    """

    Looks for FaunaDB types in a JSON object and converts to them if possible.
    """

    if "@ref" in dct:
        ref = dct["@ref"]

        if (not "collection" in ref) and (not "database" in ref):
            return Native.from_name(ref["id"])

        return Ref(ref["id"], ref.get("collection"), ref.get("database"))

    if "@obj" in dct:
        return dct["@obj"]

    if "@set" in dct:
        return SetRef(dct["@set"])

    if "@query" in dct:
        return Query(dct["@query"])

    if "@ts" in dct:
        return FaunaTime(dct["@ts"])

    if "@date" in dct:
        return parse_date(dct["@date"]).date()

    if "@bytes" in dct:
        return bytearray(urlsafe_b64decode(dct["@bytes"].encode()))
    return dct


def to_json(dct, pretty=True, sort_keys=True):
    """

    Opposite of parse_json.

    Converts a :any`_Expr` into a request body, calling :any:`to_fauna_json`.
    """
    if pretty:
        return dumps(
            dct, cls=FaunaJSONEncoder, sort_keys=True, indent=4, separators=(", ", ": ")
        )

    return dumps(dct, cls=FaunaJSONEncoder, sort_keys=sort_keys, separators=(",", ":"))


class FaunaJSONEncoder(JSONEncoder):
    """
    Encodes objects to Fauna-compatible JSON format.

    Methods:
    -------
    default(obj):
        Encodes the given object in Fauna-compatible JSON format.

    Attributes:
    ----------
    No public attributes.
    """

    @override
    def default(self, obj):
        """
        Encodes the given object in Fauna-compatible JSON format.

        Parameters:
        ----------
        obj: any
            The object to be encoded.

        Returns:
        -------
        dict
            The encoded object in Fauna-compatible JSON format.
        """

        if isinstance(obj, (Ref, SetRef, FaunaTime, Query)):
            return obj.to_fauna_json()
        if isinstance(obj, Expr):
            return obj.to_fauna_json()
        elif isinstance(obj, datetime):
            return FaunaTime(obj).to_fauna_json()
        elif isinstance(obj, date):
            return {"@date": obj.isoformat()}
        elif isinstance(obj, (bytes, bytearray)):
            return {"@bytes": urlsafe_b64encode(obj).decode("utf-8")}
        elif isinstance(obj, BaseModel):
            return obj.dict()
        elif isinstance(obj, Request):
            # Swagger UI
            if obj.content_type in (
                "application/json",
                "application/x-www-form-urlencoded",
            ):
                data = parse_json_or_none(obj.content.read_nowait().decode())
                if data:
                    return {
                        "method": obj.method,
                        "path": obj.path,
                        "headers": dict(obj.headers),
                        "body": data,
                    }
            elif obj.content_type == "multipart/form-data":
                data = {}
                for k, v in asyncio.run(obj.post()):
                    _v = None
                    try:
                        _v = loads(v)
                    except ValueError:
                        _v = "file"
                    data[k] = v
            else:
                data = None
                return {
                    "method": obj.method,
                    "query_params": dict(obj.query),
                    "path_params": dict(obj.match_info),
                    "headers": dict(obj.headers),
                    "body": data,
                }


class JSONModel(BaseModel):
    """
    A model class for JSON serialization and deserialization.

    Methods:
    -------
    to_dict(**kwargs):
        Returns the model's attributes as a dictionary.

    to_json(**kwargs) -> str:
        Returns the model's attributes as a JSON string.

    dict(**kwargs):
        Alias for to_dict() method.

    json(**kwargs) -> str:
        Alias for to_json() method.

    Attributes:
    ----------
    No public attributes.
    """

    def to_dict(self, **kwargs):
        """
        Returns the model's attributes as a dictionary.

        Parameters:
        ----------
        **kwargs:
            Optional keyword arguments that can be passed to BaseModel.dict() method.

        Returns:
        -------
        dict
            The model's attributes as a dictionary.
        """

        return parse_json_or_none(self.to_json(**kwargs))

    def to_json(self, **kwargs) -> str:
        """
        Returns the model's attributes as a JSON string.

        Parameters:
        ----------
        **kwargs:
            Optional keyword arguments that can be passed to BaseModel.dict() method.

        Returns:
        -------
        str
            The model's attributes as a JSON string.
        """

        return to_json(super().dict(**kwargs))

    @override
    def dict(self, **kwargs):
        """
        Alias for to_dict() method.

        Parameters:
        ----------
        **kwargs:
            Optional keyword arguments that can be passed to BaseModel.dict() method.

        Returns:
        -------
        dict
            The model's attributes as a dictionary.
        """

        return self.to_dict(**kwargs)

    @override
    def json(self, **kwargs) -> str:
        """
        Alias for to_json() method.

        Parameters:
        ----------
        **kwargs:
            Optional keyword arguments that can be passed to BaseModel.dict() method.

        Returns:
        -------
        str
            The model's attributes as a JSON string.
        """

        return self.to_json(**kwargs)



"""
This module contains functions that generate query objects,
suitable to be passed into :any:`FaunaClient.query`.

In addition to these functions, queries may contain:

* ``None``
* Numbers (``1, 2.5``)
* Strings (``"foo"``)
* Lists (``[1, "foo"]``)
* Dicts (``{"foo": "bar"}``)
* A :any:`Ref`, :any:`SetRef`, :any:`FaunaTime`, or :class:`datetime.date`.

For example: ``select("a", {"a": Ref("123", Ref("widgets", Native.CLASSES))})``.
"""

# pylint: disable=invalid-name, redefined-builtin

from types import FunctionType
from aiofauna import deprecated

# region Basic forms


def abort(msg):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/abort>`__."""
    return _fn({"abort": msg})


def ref(collection_ref, id=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/ref>`__."""
    if id is None:
        return _fn({"@ref": collection_ref})
    return _fn({"ref": collection_ref, "id": id})


@deprecated.deprecated("use collections instead")
def classes(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/classes>`__."""
    return _fn({"classes": scope})


def collections(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/collections>`__."""
    return _fn({"collections": scope})


def documents(collections):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/documents>`__."""
    return _fn({"documents": collections})


def databases(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/databases>`__."""
    return _fn({"databases": scope})


def indexes(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/indexes>`__."""
    return _fn({"indexes": scope})


def functions(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/functions>`__."""
    return _fn({"functions": scope})


def roles(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/roles>`__."""
    return _fn({"roles": scope})


def access_providers(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/access_providers>`__."""
    return _fn({"access_providers": scope})


def keys(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/keys>`__."""
    return _fn({"keys": scope})


def tokens(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tokens>`__."""
    return _fn({"tokens": scope})


def credentials(scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/credentials>`__."""
    return _fn({"credentials": scope})


def at(timestamp, expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/at>`__."""
    return _fn({"at": timestamp, "expr": expr})


class LetBindings:
    def __init__(self, bindings):
        self._bindings = bindings

    def in_(self, inExpr):
        return _fn({"let": self._bindings, "in": inExpr})


def let(*args, **kwargs):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/let>`__."""
    if kwargs:
        return LetBindings([_fn({k: v}) for k, v in kwargs.items()])
    else:
        bindings = [_fn({k: v}) for k, v in args[0].items()]
        inExpr = args[1]
        return _fn({"let": bindings, "in": inExpr})


def var(var_name):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/var>`__."""
    return _fn({"var": var_name})


@deprecated.deprecated("use if_ instead")
def ifExpr(condition, then, else_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/if>`__."""
    return if_(condition, then, else_)


def if_(condition, then, else_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/if>`__."""
    return _fn({"if": condition, "then": then, "else": else_})


def do(*expressions):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/do>`__."""
    return _fn({"do": expressions})


def lambda_query(func):
    """
    See the `docs <https://app.fauna.com/documentation/reference/queryapi#basic-forms>`__.
    This form generates :any:`var` objects for you, and is called like::

        query.lambda_query(lambda a: query.add(a, a))
        # Produces: {
        #  "lambda": "a",
        #  "expr": {"add": ({"var": "a"}, {"var": "a"})}
        # }

    You usually don't need to call this directly as lambdas in queries will be converted for you.
    For example: ``query.map_(lambda a: query.add(a, 1), collection)``.

    You can also use :any:`lambda_` directly.

    :param func:
      Takes one or more :any:`var` expressions and uses them to construct an expression.
      If this has more than one argument, the lambda destructures an array argument.
      (To destructure single-element arrays use :any:`lambda_`.)
    """

    vars = func.__code__.co_varnames
    n_args = len(vars)

    if n_args == 0:
        raise ValueError("Function must take at least 1 argument.")
    elif n_args == 1:
        v = vars[0]
        return lambda_(v, func(var(v)))
    else:
        return lambda_(vars, func(*[var(v) for v in vars]))


@deprecated.deprecated("use lambda_ instead")
def lambdaExpr(var_name_or_pattern, expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/lambda>`__."""
    return lambda_(var_name_or_pattern, expr)


def lambda_(var_name_or_pattern, expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/lambda>`__."""
    return _fn({"lambda": var_name_or_pattern, "expr": expr})


def call(ref_, *arguments):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/call>`__."""
    return _fn({"call": ref_, "arguments": _varargs(arguments)})


def query(_lambda):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/query>`__."""
    if isinstance(_lambda, FunctionType):
        _lambda = lambda_query(_lambda)
    return _fn({"query": _lambda})


# endregion

# region Collection functions


@deprecated.deprecated("use map_ instead")
def mapExpr(expr, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/map>`__."""
    return map_(expr, collection)


def map_(expr, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/map>`__."""
    return _fn({"map": expr, "collection": collection})


def foreach(expr, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/foreach>`__."""
    return _fn({"foreach": expr, "collection": collection})


@deprecated.deprecated("use filter_ instead")
def filterExpr(expr, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/filter>`__."""
    return filter_(expr, collection)


def filter_(expr, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/filter>`__."""
    return _fn({"filter": expr, "collection": collection})


def take(number, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/take>`__."""
    return _fn({"take": number, "collection": collection})


def drop(number, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/drop>`__."""
    return _fn({"drop": number, "collection": collection})


def prepend(elements, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/prepend>`__."""
    return _fn({"prepend": elements, "collection": collection})


def append(elements, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/append>`__."""
    return _fn({"append": elements, "collection": collection})


def is_empty(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isempty>`__."""
    return _fn({"is_empty": collection})


def is_nonempty(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isnonempty>`__."""
    return _fn({"is_nonempty": collection})


# endregion

# region Type functions


def is_number(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isnumber>`__."""
    return _fn({"is_number": expr})


def is_double(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isdouble>`__."""
    return _fn({"is_double": expr})


def is_integer(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isinteger>`__."""
    return _fn({"is_integer": expr})


def is_boolean(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isboolean>`__."""
    return _fn({"is_boolean": expr})


def is_null(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isnull>`__."""
    return _fn({"is_null": expr})


def is_bytes(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isbytes>`__."""
    return _fn({"is_bytes": expr})


def is_timestamp(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/istimestamp>`__."""
    return _fn({"is_timestamp": expr})


def is_date(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isdate>`__."""
    return _fn({"is_date": expr})


def is_string(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isstring>`__."""
    return _fn({"is_string": expr})


def is_array(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isarray>`__."""
    return _fn({"is_array": expr})


def is_object(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isobject>`__."""
    return _fn({"is_object": expr})


def is_ref(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isref>`__."""
    return _fn({"is_ref": expr})


def is_set(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isset>`__."""
    return _fn({"is_set": expr})


def is_doc(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isdoc>`__."""
    return _fn({"is_doc": expr})


def is_lambda(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/islambda>`__."""
    return _fn({"is_lambda": expr})


def is_collection(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/iscollection>`__."""
    return _fn({"is_collection": expr})


def is_database(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isdatabase>`__."""
    return _fn({"is_database": expr})


def is_index(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isindex>`__."""
    return _fn({"is_index": expr})


def is_function(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isfunction>`__."""
    return _fn({"is_function": expr})


def is_key(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/iskey>`__."""
    return _fn({"is_key": expr})


def is_token(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/istoken>`__."""
    return _fn({"is_token": expr})


def is_credentials(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/iscredentials>`__."""
    return _fn({"is_credentials": expr})


def is_role(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/isrole>`__."""
    return _fn({"is_role": expr})


# endregion

# region Read functions


def get(ref_, ts=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/get>`__."""
    return _params({"get": ref_}, {"ts": ts})


def key_from_secret(secret):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/keyfromsecret>`__."""
    return _fn({"key_from_secret": secret})


def paginate(
    set, size=None, ts=None, after=None, before=None, events=None, sources=None
):
    """
    See the `docs <https://app.fauna.com/documentation/reference/queryapi#read-functions>`__.
    You may want to convert the result of this to a :any:`Page`.
    """
    # pylint: disable=too-many-arguments
    # pylint: disable=redefined-outer-name
    opts = {
        "size": size,
        "ts": ts,
        "after": after,
        "before": before,
        "events": events,
        "sources": sources,
    }
    return _params({"paginate": set}, opts)


def exists(ref_, ts=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/exists>`__."""
    return _params({"exists": ref_}, {"ts": ts})


# endregion

# region Write functions


def create(collection_ref, params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/create>`__."""
    return _fn({"create": collection_ref, "params": params})


def update(ref_, params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/update>`__."""
    return _fn({"update": ref_, "params": params})


def replace(ref_, params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/replace>`__."""
    return _fn({"replace": ref_, "params": params})


def delete(ref_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/delete>`__."""
    return _fn({"delete": ref_})


def insert(ref_, ts, action, params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/insert>`__."""
    return _fn({"insert": ref_, "ts": ts, "action": action, "params": params})


def remove(ref_, ts, action):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/remove>`__."""
    return _fn({"remove": ref_, "ts": ts, "action": action})


@deprecated.deprecated("use create_collection instead")
def create_class(class_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/create_class>`__."""
    return _fn({"create_class": class_params})


def create_collection(collection_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createcollection>`__."""
    return _fn({"create_collection": collection_params})


def create_database(db_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createdatabase>`__."""
    return _fn({"create_database": db_params})


def create_index(index_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createindex>`__."""
    return _fn({"create_index": index_params})


def create_function(func_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createfunction>`__."""
    return _fn({"create_function": func_params})


def create_role(func_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createrole>`__."""
    return _fn({"create_role": func_params})


def create_access_provider(provider_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createaccessprovider>`__."""
    return _fn({"create_access_provider": provider_params})


def move_database(from_, to):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/movedatabase>`__."""
    return _fn({"move_database": from_, "to": to})


def create_key(key_params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/createkey>`__."""
    return _fn({"create_key": key_params})


# endregion

# region Sets


def singleton(ref_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/singleton>`__."""
    return _fn({"singleton": ref_})


def events(ref_set):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/events>`__."""
    return _fn({"events": ref_set})


def match(index, *terms):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/match>`__."""
    # pylint: disable=redefined-outer-name
    m = {"match": index}

    if len(terms) >= 1:
        m["terms"] = _varargs(terms)

    return _fn(m)


def reverse(set_array_or_page):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/reverse>`__."""
    return _fn({"reverse": set_array_or_page})


def merge(merge, with_, lambda_=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/merge>`__."""
    return _params({"merge": merge, "with": with_}, {"lambda": lambda_})


def union(*sets):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/union>`__."""
    return _fn({"union": _varargs(sets)})


def reduce(lambda_, initial, collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/reduce>`__."""
    return _fn({"reduce": lambda_, "initial": initial, "collection": collection})


def intersection(*sets):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/intersection>`__."""
    return _fn({"intersection": _varargs(sets)})


def difference(*sets):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/difference>`__."""
    return _fn({"difference": _varargs(sets)})


def distinct(set):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/distinct>`__."""
    return _fn({"distinct": set})


def join(source, target):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/join>`__."""
    return _fn({"join": source, "with": target})


def range(set, from_, to):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/range>`__."""
    return _fn({"range": set, "from": from_, "to": to})


# endregion

# region Authentication


def login(ref_, params):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/login>`__."""
    return _fn({"login": ref_, "params": params})


def logout(delete_tokens):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/logout>`__."""
    return _fn({"logout": delete_tokens})


def identify(ref_, password):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/identify>`__."""
    return _fn({"identify": ref_, "password": password})


@deprecated.deprecated("Use `current_identity` instead")
def identity():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/identity>`__."""
    return _fn({"identity": None})


def current_identity():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/current_identity>`__."""
    return _fn({"current_identity": None})


def has_current_identity():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/has_current_identity>`__."""
    return _fn({"has_current_identity": None})


def current_token():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/current_token>`__."""
    return _fn({"current_token": None})


def has_current_token():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/has_current_token>`__."""
    return _fn({"has_current_token": None})


def has_identity():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/hasidentity>`__."""
    return _fn({"has_identity": None})


# endregion

# region String functions


def format(string, *values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/format>`__."""
    return _fn({"format": string, "values": _varargs(values)})


def concat(strings, separator=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/concat>`__."""
    return _params({"concat": strings}, {"separator": separator})


def casefold(string, normalizer=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/casefold>`__."""
    return _params({"casefold": string}, {"normalizer": normalizer})


def starts_with(value, search):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/startswith>`__."""
    return _fn({"startswith": value, "search": search})


def ends_with(value, search):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/endswith>`__."""
    return _fn({"endswith": value, "search": search})


def contains_str(value, search):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/containsstr>`__."""
    return _fn({"containsstr": value, "search": search})


def contains_str_regex(value, pattern):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/containsstrregex>`__."""
    return _fn({"containsstrregex": value, "pattern": pattern})


def regex_escape(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/regexescape>`__."""
    return _fn({"regexescape": value})


def ngram(terms, min=None, max=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/ngram>`__."""
    return _params({"ngram": terms}, {"min": min, "max": max})


def find_str(value, find, start=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/findstr>`__."""
    return _params({"findstr": value, "find": find}, {"start": start})


def find_str_regex(value, pattern, start=None, numResults=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/findstrregex>`__."""
    return _params(
        {"findstrregex": value, "pattern": pattern},
        {"start": start, "num_results": numResults},
    )


def replace_str(value, find, replace):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/replacestr>`__."""
    return _fn({"replacestr": value, "find": find, "replace": replace})


def replace_str_regex(value, pattern, replace, first=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/replacestrregex>`__."""
    return _params(
        {"replacestrregex": value, "pattern": pattern, "replace": replace},
        {"first": first},
    )


def length(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/length>`__."""
    return _fn({"length": value})


def lowercase(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/lowercase>`__."""
    return _fn({"lowercase": value})


def uppercase(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/uppercase>`__."""
    return _fn({"uppercase": value})


def titlecase(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/titlecase>`__."""
    return _fn({"titlecase": value})


def trim(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/trim>`__."""
    return _fn({"trim": value})


def ltrim(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/ltrim>`__."""
    return _fn({"ltrim": value})


def rtrim(value):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/rtrim>`__."""
    return _fn({"rtrim": value})


def space(count):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/space>`__."""
    return _fn({"space": count})


def substring(value, start, length=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/substring>`__."""
    return _params({"substring": value, "start": start}, {"length": length})


def repeat(value, number=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/repeat>`__."""
    return _params({"repeat": value}, {"number": number})


# endregion

# region Time and date functions


def time(string):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/time>`__."""
    return _fn({"time": string})


def epoch(number, unit):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/epoch>`__."""
    return _fn({"epoch": number, "unit": unit})


def now():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/date>`__."""
    return _fn({"now": None})


def date(string):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/date>`__."""
    return _fn({"date": string})


def time_add(base, offset, unit):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/timeadd>`__."""
    return _fn({"time_add": base, "offset": offset, "unit": unit})


def time_subtract(base, offset, unit):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/timesubtract>`__."""
    return _fn({"time_subtract": base, "offset": offset, "unit": unit})


def time_diff(start, finish, unit):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/timediff>`__."""
    return _fn({"time_diff": start, "other": finish, "unit": unit})


# endregion

# region Miscellaneous functions


@deprecated.deprecated("use new_id instead")
def next_id():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/nextid>`__."""
    return _fn({"next_id": None})


def new_id():
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/newid>`__."""
    return _fn({"new_id": None})


def database(db_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/database>`__."""
    return _params({"database": db_name}, {"scope": scope})


def index(index_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/index>`__."""
    return _params({"index": index_name}, {"scope": scope})


@deprecated.deprecated("use collection instead")
def classExpr(class_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/class>`__."""
    return class_(class_name, scope)


@deprecated.deprecated("use collection instead")
def class_(class_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/class>`__."""
    return _params({"class": class_name}, {"scope": scope})


def collection(collection_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/collection>`__."""
    return _params({"collection": collection_name}, {"scope": scope})


def function(fn_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/function>`__."""
    return _params({"function": fn_name}, {"scope": scope})


def role(role_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/role>`__."""
    return _params({"role": role_name}, {"scope": scope})


def access_provider(access_provider_name, scope=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/access_provider>`__."""
    return _params({"access_provider": access_provider_name}, {"scope": scope})


def equals(*values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/equals>`__."""
    return _fn({"equals": _varargs(values)})


@deprecated.deprecated("use `contains_path` instead.")
def contains(path, in_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/contains>`__."""
    return _fn({"contains": path, "in": in_})


def contains_path(path, in_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/containspath>`__."""
    return _fn({"contains_path": path, "in": in_})


def contains_field(field, in_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/containsfield>`__."""
    return _fn({"contains_field": field, "in": in_})


def contains_value(value, in_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/containsvalue>`__."""
    return _fn({"contains_value": value, "in": in_})


_NO_DEFAULT = object()


def select(path, from_, default=_NO_DEFAULT):
    """
    See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/select>`__.
    See also :py:func:`select_with_default`."""
    _dict = {"select": path, "from": from_}
    if default is not _NO_DEFAULT:
        _dict["default"] = default
    return _fn(_dict)


@deprecated.deprecated("Use `select` instead")
def select_with_default(path, from_, default):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/select>`__."""
    return _fn({"select": path, "from": from_, "default": default})


def select_all(path, from_):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/selectall>`__."""
    return _fn({"select_all": path, "from": from_})


def add(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/add>`__."""
    return _fn({"add": _varargs(numbers)})


def multiply(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/multiply>`__."""
    return _fn({"multiply": _varargs(numbers)})


def subtract(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/subtract>`__."""
    return _fn({"subtract": _varargs(numbers)})


def divide(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/divide>`__."""
    return _fn({"divide": _varargs(numbers)})


def pow(base, exp):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/pow>` __."""
    return _fn({"pow": base, "exp": exp})


def max(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/max>` __."""
    return _fn({"max": _varargs(numbers)})


def min(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/min>` __."""
    return _fn({"min": _varargs(numbers)})


def abs(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/abs>` __."""
    return _fn({"abs": num})


def trunc(num, precision=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/trunc>` __."""
    return _params({"trunc": num}, {"precision": precision})


def bitor(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/bitor>` __."""
    return _fn({"bitor": _varargs(numbers)})


def cosh(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/cosh>` __."""
    return _fn({"cosh": num})


def hypot(num, b):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/hypot>` __."""
    return _fn({"hypot": num, "b": b})


def atan(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/atan>` __."""
    return _fn({"atan": num})


def log(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/log>` __."""
    return _fn({"log": num})


def bitnot(*num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/bitnot>` __."""
    return _fn({"bitnot": _varargs(num)})


def bitxor(*num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/bitxor>` __."""
    return _fn({"bitxor": _varargs(num)})


def bitand(*num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/bitand>` __."""
    return _fn({"bitand": _varargs(num)})


def ceil(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/ceil>` __."""
    return _fn({"ceil": num})


def degrees(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/degrees>` __."""
    return _fn({"degrees": num})


def cos(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/cos>` __."""
    return _fn({"cos": num})


def acos(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/acos>` __."""
    return _fn({"acos": num})


def sqrt(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/sqrt>` __."""
    return _fn({"sqrt": num})


def tan(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tan>` __."""
    return _fn({"tan": num})


def tanh(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tanh>` __."""
    return _fn({"tanh": num})


def sin(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/sin>` __."""
    return _fn({"sin": num})


def asin(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/asin>` __."""
    return _fn({"asin": num})


def round(num, precision=None):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/round>` __."""
    return _params({"round": num}, {"precision": precision})


def radians(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/radians>` __."""
    return _fn({"radians": num})


def floor(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/floor>` __."""
    return _fn({"floor": num})


def sign(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/sign>` __."""
    return _fn({"sign": num})


def exp(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/exp>` __."""
    return _fn({"exp": num})


def ln(num):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/ln>` __."""
    return _fn({"ln": num})


def any(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/any>`__."""
    return _fn({"any": collection})


def all(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/all>`__."""
    return _fn({"all": collection})


def modulo(*numbers):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/modulo>`__."""
    return _fn({"modulo": _varargs(numbers)})


def count(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/count>`__."""
    return _fn({"count": collection})


def sum(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/sum>`__."""
    return _fn({"sum": collection})


def mean(collection):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/mean>`__."""
    return _fn({"mean": collection})


def lt(*values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/lt>`__."""
    return _fn({"lt": _varargs(values)})


def lte(*values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/lte>`__."""
    return _fn({"lte": _varargs(values)})


def gt(*values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/gt>`__."""
    return _fn({"gt": _varargs(values)})


def gte(*values):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/gte>`__."""
    return _fn({"gte": _varargs(values)})


@deprecated.deprecated("use and_ instead")
def andExpr(*booleans):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/and>`__."""
    return and_(*booleans)


def and_(*booleans):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/and>`__."""
    return _fn({"and": _varargs(booleans)})


@deprecated.deprecated("use or_ instead")
def orExpr(*booleans):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/or>`__."""
    return or_(*booleans)


def or_(*booleans):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/or>`__."""
    return _fn({"or": _varargs(booleans)})


@deprecated.deprecated("use not_ instead")
def notExpr(boolean):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/not>`__."""
    return not_(boolean)


def not_(boolean):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/not>`__."""
    return _fn({"not": boolean})


def to_string(expr):
    return _fn({"to_string": expr})


def to_array(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/toarray>`__."""
    return _fn({"to_array": expr})


def to_object(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/toobject>`__."""
    return _fn({"to_object": expr})


def to_double(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/todouble>`__."""
    return _fn({"to_double": expr})


def to_integer(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tointeger>`__."""
    return _fn({"to_integer": expr})


def to_number(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tonumber>`__."""
    return _fn({"to_number": expr})


def to_time(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/totime>`__."""
    return _fn({"to_time": expr})


def to_seconds(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/toseconds>`__."""
    return _fn({"to_seconds": expr})


def to_millis(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tomillis>`__."""
    return _fn({"to_millis": expr})


def to_micros(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/tomicros>`__."""
    return _fn({"to_micros": expr})


def day_of_month(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/dayofmonth>`__."""
    return _fn({"day_of_month": expr})


def day_of_week(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/dayofweek>`__."""
    return _fn({"day_of_week": expr})


def day_of_year(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/dayofyear>`__."""
    return _fn({"day_of_year": expr})


def year(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/year>`__."""
    return _fn({"year": expr})


def month(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/month>`__."""
    return _fn({"month": expr})


def hour(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/hour>`__."""
    return _fn({"hour": expr})


def minute(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/minute>`__."""
    return _fn({"minute": expr})


def second(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/second>`__."""
    return _fn({"second": expr})


def to_date(expr):
    """See the `docs <https://docs.fauna.com/fauna/current/api/fql/functions/todate>`__."""
    return _fn({"to_date": expr})


# endregion

# region Helpers


class Expr(object):
    """
    Used to mark values that have been wrapped.
    """

    def __init__(self, value):
        self.value = value

    def to_fauna_json(self):
        return self.value

    def __repr__(self):
        return "Expr(%s)" % repr(self.value)

    def __eq__(self, other):
        return isinstance(other, Expr) and self.value == other.value


def _wrap(value):
    if isinstance(value, Expr):
        return value
    elif isinstance(value, FunctionType):
        return lambda_query(value)
    elif isinstance(value, dict):
        return Expr({"object": _wrap_values(value)})
    elif isinstance(value, (list, tuple)):
        return Expr([_wrap(sub_value) for sub_value in value])
    return value


def _wrap_values(dct):
    return {key: _wrap(val) for key, val in dct.items()}


def _fn(dct):
    return Expr(_wrap_values(dct))


def _params(main_params, optional_params):
    """Hash of query arguments with None values removed."""
    for key, val in optional_params.items():
        if val is not None:
            main_params[key] = val
    return _fn(main_params)


def _varargs(values):
    """
    Called on ``*args`` arguments.
    This ensures that a single value passed is not put in an array, so
    ``query.add([1, 2])`` will work as well as ``query.add(1, 2)``.
    """
    return values[0] if len(values) == 1 else values


# endregion



"""REST API Module with automatic OpenAPI generation."""
from __future__ import annotations
import json
import base64
import asyncio
from json import JSONDecoder
from typing import Any, Dict, List, Union
from inspect import signature
from uuid import UUID
from datetime import datetime, date, time
from functools import singledispatch
from enum import Enum
from dataclasses import dataclass
import uuid
from aiohttp.web import Application, Request, Response, json_response
from aiohttp.web_request import FileField
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from multidict import CIMultiDict   
from .json import FaunaJSONEncoder, parse_json_or_none
from .odm import AsyncFaunaModel



class UploadFile(BaseModel):
    """
    File Upload Model
    """

    filename: str
    content_type: str
    data: bytes
    size: int
    headers: Dict[str, str]

        
            
def jsonable_encoder(
    obj: Any,
    *,
    include: List[str] = [],
    exclude: List[str] = [],
    by_alias: bool = False,
    skip_defaults: bool = False,
    custom_encoder: Any = None,
) -> Any:
    """
    Convert any object to a JSON-serializable object.

    This function is used by Aiofauna to convert objects to JSON-serializable objects.

    It supports all the types supported by the standard json library, plus:

    * datetime.datetime
    * datetime.date
    * datetime.time
    * uuid.UUID
    * enum.Enum
    * pydantic.BaseModel
    """

    if custom_encoder is None:
        custom_encoder = FaunaJSONEncoder

    if obj is str:
        return "string"
    if obj is int or obj is float:
        return "integer"
    if obj is bool:
        return "boolean"
    if obj is None:
        return "null"
    if obj is list:
        return "array"
    if obj is dict:
        return "object"
    if obj is bytes:
        return "binary"
    if obj is datetime:
        return "date-time"
    if obj is date:
        return "date"
    if obj is time:
        return "time"
    if obj is UUID:
        return "uuid"
    if obj is Enum:
        return "enum"
    if isinstance(obj, (str, int, float, bool, type(None))):
        return obj
    if isinstance(obj, (list, tuple, set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, dict):
        return {
            jsonable_encoder(
                k,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            ): jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for k, v in obj.items()
        }
    if isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    if isinstance(obj, (set, frozenset)):
        return [
            jsonable_encoder(
                v,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                custom_encoder=custom_encoder,
            )
            for v in obj
        ]
    if isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, UUID):
        return str(obj)
    return custom_encoder().default(obj)

def extract_params(params: dict, path: str):
    """
    FastAPIesque function to extract the parameters of a function outside of the request context.
    """

    open_api_params = {}

    for name, param in params.items():
        type_ = param.annotation

        if type_ in (str, int, float, bool) and name:
            if f"{{{name}}}" in path:
                param_location = "path"
            else:
                param_location = "query"

            open_api_params[name] = {
                "in": param_location,
                "name": name,
                "required": True,
                "schema": {"type": type_, "default": param.default, "required": True},
            }

        elif type_ in [UploadFile, FileField]:
            open_api_params[name] = {
        "in": "formData",
        "name": name,
        "required": True,
        "schema": {"type": "file", "format": "binary"},
        "consumes": ["multipart/form-data"],
        "headers": {
            "Content-Type": {
                "type": "string",
                "default": "multipart/form-data"
            }
        }
    }
        
        
        elif issubclass(type_, (BaseModel, AsyncFaunaModel)) and type_ != UploadFile:
            open_api_params[name] = {
                "in": "body",
                "name": name,
                "required": True,
                "schema": type_.schema(),
            }

        
        else:
            continue

    return open_api_params


def update_open_api(
    open_api: Dict[str, Any],
    path: str,
    method: str,
    func: Any,
    open_api_params: Dict[str, Any],
) -> None:
    if path in ["/openapi.json", "/docs"]:
        return

    _scalars = []
    _body = None
    _is_file_upload = False

    for param in open_api_params.values():
        if isinstance(param["schema"], dict):
            if "type" in param["schema"] and param["schema"]["type"] != "object":
                _scalars.append(param)
            else:
                _body = {
                    "content": {"application/json": {"schema": param["schema"]}}
                }
        elif param["in"] == "formData" and param["schema"]["type"] == "file":
            _is_file_upload = True
            _scalars.append(param)
        else:
            continue

    if _body:
        open_api["paths"].setdefault(path, {})[method.lower()] = {
        "summary": func.__name__,
        "description": func.__doc__,
        "parameters": _scalars,
        "requestBody": _body if not _is_file_upload else {
            "content": {
                "multipart/form-data": {
                    "schema": {
                        "properties": {
                            "file": {
                                "type": "array",
                                "items": {
                                    "type": "string",
                                    "format": "binary"
                                }
                            }
                        }
                    }
                }
            }
        },
        "responses": {"200": {"description": "OK"}},
    }
        open_api["components"]["schemas"].update(param["schema"])  # type: ignore
    else:
        open_api["paths"].setdefault(path, {})[method.lower()] = {
            "summary": func.__name__,
            "description": func.__doc__,
            "parameters": _scalars,
            "responses": {"200": {"description": "OK"}},
        }
        
async def inject_signature(request: Request, params: dict):
    """
    FastAPIesque function to shape the signature of a function to the request.
    """
    args_to_apply = {}

    for name, param in params.items():
        annotation = param.annotation
        if annotation in (str, int, float, bool) and name in request.match_info:
            args_to_apply[name] = request.match_info[name]
        elif annotation in (str, int, float, bool) and name in request.query:
            args_to_apply[name] = annotation(request.query[name])
        elif annotation in [FileField, UploadFile]:
                headers = dict(request.headers)
                new_headers = CIMultiDict(**headers, **{"content-type": "multipart/form-data"})
                new_request = request.clone(headers=new_headers)
                print(new_request.has_body)
                args_to_apply[name] = new_request
        elif issubclass(annotation, BaseModel):
            data = await request.json(loads=JSONDecoder().decode)
            if isinstance(data, Union[str,bytes]):
                data = json.loads(data, object_hook=parse_json_or_none)
            args_to_apply[name] = annotation(**data)
        else:
            continue
    return args_to_apply


@singledispatch
def do_response(response: Any) -> Response:
    """
    Flask-esque function to make a response from a function.
    """
    return response


@do_response.register(BaseModel)
def _(response: BaseModel) -> Response:
    return Response(
        status=200,
        body=json.dumps(response.dict()),
        content_type="application/json",
    )


@do_response.register(AsyncFaunaModel)
def _(response: AsyncFaunaModel) -> Response:
    response.ref = str(response.ref)
    return Response(
        status=200,
        body=json.dumps(response.dict()),
        content_type="application/json",
    )


@do_response.register(dict)
def _(response: dict) -> Response:
    return Response(
        status=200,
        body=json.dumps(response),
        content_type="application/json",
    )


@do_response.register(str)
def _(response: str) -> Response:
    return Response(
        status=200,
        body=response.encode(),
        content_type="text/plain",
    )


@do_response.register(bytes)
def _(response: bytes) -> Response:
    return Response(
        status=200,
        body=response,
        content_type="application/octet-stream",
    )


@do_response.register(int)
def _(response: int) -> Response:
    return Response(
        status=200,
        body=str(response).encode(),
        content_type="text/plain",
    )


@do_response.register(float)
def _(response: float) -> Response:
    return Response(
        status=200,
        body=str(response).encode(),
        content_type="text/plain",
    )


@do_response.register(bool)
def _(response: bool) -> Response:
    return Response(
        status=200,
        body=str(response).encode(),
        content_type="text/plain",
    )


@do_response.register(list)
def _(
    response: List[Union[AsyncFaunaModel, BaseModel, dict, str, int, float]]
) -> Response:
    processed_response = []

    for item in response:
        if isinstance(item, (BaseModel, AsyncFaunaModel)):
            processed_response.append(item.dict())
        else:
            processed_response.append(item)

    return Response(
        status=200,
        body=json.dumps(processed_response),
        content_type="application/json",
    )


class Api(Application):
    """

    Api class to create a fastapi-like api.

    """

    def __init__(self, *args, title: str = "AioFauna API", **kwargs):
        super().__init__(*args, **kwargs)
        self.title = title
        self.openapi = {
            "openapi": "3.0.0",
            "info": {"title": self.title, "version": "1.0.0"},
            "paths": {},
            "components": {"schemas": {}},
            "description": "AioFauna API",
        }
        self._route_open_api_params = {}

        @self.get("/openapi.json")
        async def openapi():
            response = jsonable_encoder(self.openapi)
            return json_response(response)

        @self.get("/docs")
        async def docs():
            html = f"""<!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <title>{self.title}</title>
                <link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui.css" >
                <link rel="icon" type="image/png" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/favicon-32x32.png" sizes="32x32" />
                <link rel="icon" type="image/png" href="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/favicon-16x16.png" sizes="16x16" />
                <style>
                html
                {{
                    box-sizing: border-box;
                    overflow: -moz-scrollbars-vertical;
                    overflow-y: scroll;
                }}
                
                .swagger-ui .topbar
                {{
                    display: none;
                }}
                    
                    
                *,
                *:before,
                *:after
                {{
                    box-sizing: inherit;
                }}

                body
                {{
                    margin:0;
                    background: #fafafa;
                }}
                </style>
            </head>

            <body>
                <div id="swagger-ui"></div>

                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui-bundle.js"> </script>
                <script src="https://cdn.jsdelivr.net/npm/swagger-ui-dist@3.20.3/swagger-ui-standalone-preset.js"> </script>
                <script>
                window.onload = function() {{
                const ui = SwaggerUIBundle({{
                    url: "/openapi.json",
                    dom_id: '#swagger-ui',
                    deepLinking: true,
                    presets: [
                    SwaggerUIBundle.presets.apis,
                    SwaggerUIStandalonePreset
                    ],
                    plugins: [
                    SwaggerUIBundle.plugins.DownloadUrl
                    ],
                    layout: "StandaloneLayout"
                }})
                window.ui = ui
                }}
            </script>
            </body>
            </html>
            """
            return Response(body=html.encode(), content_type="text/html")

    def document(self, path: str, method: str):
        """

        Decorator to document a function.

        """

        def decorator(func):
            sig = signature(func)
            params = sig.parameters
            open_api_params = extract_params(params.copy(), path)
            self._route_open_api_params[(path, method)] = open_api_params
            update_open_api(self.openapi, path, method, func, open_api_params)

            async def wrapper(*args, **kwargs):
                request: Request = args[0]
                args = args[1:]
                args_to_apply = await inject_signature(request, params.copy())
                definitive_args = {}
                for name, param in params.items():
                    if name in args_to_apply:
                        definitive_args[name] = args_to_apply[name]
                    elif param.default is not param.empty:
                        definitive_args[name] = param.default
                    else:
                        raise ValueError(
                            f"Missing parameter {name} for {func.__name__}"
                        )
                if asyncio.iscoroutinefunction(func):
                    response = await func(*args, **kwargs, **definitive_args)
                else:
                    response = func(*args, **kwargs, **definitive_args)
                return do_response(response)

            wrapper._handler = func
            return wrapper

        return decorator

    def get(self, path: str, **kwargs):
        """GET decorator"""
        def decorator(func):
            self.router.add_get(path, self.document(path, "GET")(func), **kwargs)
            return func

        return decorator

    def post(self, path: str, **kwargs):
        """POST decorator"""
        def decorator(func):
            self.router.add_post(path, self.document(path, "POST")(func), **kwargs)
            return func

        return decorator

    def put(self, path: str, **kwargs):
        """PUT decorator"""
        def decorator(func):
            self.router.add_put(path, self.document(path, "PUT")(func), **kwargs)
            return func

        return decorator

    def delete(self, path: str, **kwargs):
        """DELETE decorator"""
        def decorator(func):
            self.router.add_delete(path, self.document(path, "DELETE")(func), **kwargs)
            return func

        return decorator

    def patch(self, path: str, **kwargs):
        """PATCH decorator"""
        def decorator(func):
            self.router.add_patch(path, self.document(path, "PATCH")(func), **kwargs)
            return func

        return decorator

    def head(self, path: str, **kwargs):
        """HEAD decorator"""
        def decorator(func):
            self.router.add_head(path, self.document(path, "HEAD")(func), **kwargs)
            return func

        return decorator

    def options(self, path: str, **kwargs):
        """OPTIONS decorator"""
        def decorator(func):
            self.router.add_options(
                path, self.document(path, "OPTIONS")(func), **kwargs
            )
            return func

        return decorator

    def on_event(self, event: str):
        """On event handler"""
        def decorator(func):
            if event not in ["startup", "shutdown"]:
                raise ValueError("Event must be startup or shutdown")
            elif event == "startup":
                self.on_startup.append(func)
            else:
                self.on_shutdown.append(func)
            return func

        return decorator



"""
Flaskaesque helper functions for aiohttp.
"""
import re
from markdown import markdown  # pylint: disable=import-error
from aiohttp import web
from aiofauna import Response
from pygments import highlight
from pygments.formatters import HtmlFormatter  # pylint: disable=no-name-in-module
from pygments.lexers import get_lexer_by_name
from jinja2 import Environment, FileSystemLoader, select_autoescape

MD_FMT_CDN = """<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/github-markdown-css/github-markdown.min.css">"""  # pylint: disable=line-too-long
CODE_FMT_CDN = """<link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/pygments-css@1.0.0/github.min.css">"""  # pylint: disable=line-too-long

html_formatter = HtmlFormatter(style="default")

jinja_env = Environment(
    loader=FileSystemLoader("templates"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
    extensions=["jinja2.ext.do"],
)

md_env = Environment(
    loader=FileSystemLoader("markdown"),
    autoescape=select_autoescape(["html", "xml"]),
    trim_blocks=True,
    lstrip_blocks=True,
)


def markdown_it(md_string: str):
    """
    Render a markdown file to HTML.

    Args:
        md_string (str): The markdown file to render.

    Returns:
        str: The rendered HTML.
    """

    template = md_env.get_template(md_string)
    text = template.render()
    code = re.findall(r"```[a-z]*\n[\s\S]*?\n```", text)
    for code_ in code:
        lang = re.findall(r"```([a-z]*)\n", code_)[0]
        lexer = get_lexer_by_name(lang)
        code = re.findall(r"```[a-z]*\n([\s\S]*?)\n```", code_)[0]
        code = highlight(code, lexer, html_formatter)
        text = text.replace(code_, code)
    text = markdown(text, extensions=["fenced_code"])
    text = f"""
    <html>
        <head>
            <meta charset="utf-8">
            {MD_FMT_CDN}
            {CODE_FMT_CDN}
        </head>
        <style>
        .container {{
            max-width: 800px;
            margin: 0 auto;
            padding: 20px;
        }}
        h1, h2, h3, h4, h5, h6 {{
            margin-top: 40px;
            text-align: center;
            font-weight: 600;
            font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Helvetica, Arial, sans-serif;
        }}
        </style>
        <body class="markdown-body">
        <main class="container">
            {text}
        </main>
        </body>
    </html>
    """
    return Response(text=text, content_type="text/html")


def redirect(url):
    """
    Create an HTTP redirect response.

    Args:
        url (str): The URL to redirect to.

    Returns:
        web.HTTPFound: An HTTP redirect response with the specified URL.
    """

    return web.HTTPFound(url)


def make_response(data, headers=None, status=200):
    """
    Create an HTTP response with the given data, headers, and status code.

    Args:
        data (str): The response body data.
        headers (dict, optional): A dictionary of headers to include in the response. Defaults to None.
        status (int, optional): The HTTP status code for the response. Defaults to 200.

    Returns:
        web.Response: An HTTP response with the specified data, headers, and status code.
    """
    return web.Response(body=data, headers=headers, status=status)


def render_template(template_name, **kwargs):
    """
    Render an HTML template with the given context variables and return an HTTP response.

    Args
        template_name (str): The name of the template file.
        **kwargs: Keyword arguments representing context variables to be passed to the template.

    Returns:
        web.Response: An HTTP response with the rendered template as its body and content type set to "text/html".
    """
    template = jinja_env.get_template(template_name).render(**kwargs)
    return web.Response(body=template.encode(), content_type="text/html")



"""Lightweight ORM to perform simple CRUD operations on FaunaDB collections and provision indexes


   the fauna query object is available also within the class for further customization
"""


from __future__ import annotations


import logging

import os

import asyncio


from typing import List, Optional, Any, Callable, Union

try:
    import query as q  # type: ignore
except ImportError:
    from . import query as q

from .errors import AioFaunaException


from .json import JSONModel  # pylint: disable=no-name-in-module


from .client import AsyncFaunaClient


from pydantic import BaseModel  # pylint: disable=no-name-in-module


from dotenv import load_dotenv


load_dotenv()


class Fql(BaseModel):
    field: str

    operator: str

    value: Any


class AsyncFaunaModel(JSONModel):
    """

    FaunaDB Base Model
    """

    ref: Optional[str] = None

    ts: Optional[str] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

        for field in self.__fields__.values():
            try:
                one_of = field.field_info.extra.get("oneOf")

                if isinstance(one_of, list):
                    if data.get(field.name) not in one_of:
                        raise ValueError(f"{field.name} must be one of {one_of}")

            except Exception:  # pylint: disable=broad-except
                continue

    @classmethod
    def client(cls) -> AsyncFaunaClient:
        fauna_secret = os.getenv("FAUNA_SECRET")

        return AsyncFaunaClient(secret=fauna_secret)

    @classmethod
    def q(cls) -> Callable:
        return cls.client().query

    @classmethod
    async def provision(cls) -> bool:
        _q = cls.q()
        try:
            if not await _q(q.exists(q.collection(cls.__name__.lower()))):
                await _q(q.create_collection({"name": cls.__name__.lower()}))

                print("Created collection %s", cls.__name__.lower())

            if not await _q(q.exists(q.index(cls.__name__.lower()))):
                await _q(
                    q.create_index(
                        {
                            "name": cls.__name__.lower(),
                            "source": q.collection(cls.__name__.lower()),
                        }
                    )
                )

                print(f"Created index {cls.__name__.lower()}")

            for field in cls.__fields__.values():
                if field.field_info.extra.get("unique"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}_unique"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}_unique",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                    "unique": True,
                                }
                            )
                        )

                        print(
                            cls.__name__.lower(),
                            field.name,
                        )

                    print(
                        "Created unique index %s_%s",
                        cls.__name__.lower(),
                        field.name,
                    )

                    continue

                if field.field_info.extra.get("index"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                }
                            )
                        )

                        print("Created index %s_%s", cls.__name__.lower(), field.name)
                        continue

            return True

        except AioFaunaException as exc:
            logging.error(exc)

            return False

    @classmethod
    async def exists(cls, ref: str) -> bool:
        """


        Checks if a document exists in FaunaDB.



        Parameters:
        -----------
        ref: int


            The reference ID of the document to check.



        Returns:
        --------


        bool:


            True if the document exists, False otherwise.
        """

        try:
            return await cls.q()(
                q.exists(q.ref(q.collection(cls.__name__.lower()), ref))
            )

        except AioFaunaException as exc:
            logging.error(exc)

            return False

    @classmethod
    async def find_unique(cls, field: str, value: Any) -> AsyncFaunaModel:
        """


        Finds a document in FaunaDB by a unique field.



        Parameters:
        -----------
        field: str


            The name of the field to search.


        value: Any


            The value to search for.



        Returns:
        --------


        Union[AsyncFaunaModel,BaseModel]:


            An instance of the model if found, None otherwise.
        """

        try:
            data = await cls.q()(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:
            logging.error(exc)

            return None  # type: ignore # pylint: disable=unreachable

    @classmethod
    async def find_many(cls, field: str, value: Any) -> List[AsyncFaunaModel]:
        """


        Finds documents in FaunaDB by a field.



        Parameters:
        -----------
        field: str


            The name of the field to search.


        value: Any


            The value to search for.



        Returns:
        --------


        List[AsyncFaunaModel]:


            A list of instances of the model if found, None otherwise.
        """

        try:
            _q = cls.q()

            refs = (
                await _q(
                    q.paginate(
                        q.match(q.index(f"{cls.__name__.lower()}_{field}"), value)
                    )
                )
            )["data"]

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:
            logging.error(exc)

            return []

    @classmethod
    async def find(cls, ref: str) -> AsyncFaunaModel:
        """


        Finds a document in FaunaDB by its ID.



        Parameters:
        -----------
        ref: int


            The reference ID of the document to find.



        Returns:
        --------


        Union[AsyncFaunaModel,BaseModel]:


            An instance of the model if found, None otherwise.
        """

        try:
            data = await cls.q()(q.get(q.ref(q.collection(cls.__name__.lower()), ref)))
            return cls(
                **{
                    **data["data"],
                    "ref": data["ref"]["@ref"]["id"],
                    "ts": data["ts"] / 1000,
                }
            )

        except AioFaunaException as exc:
            logging.error(exc)

            return None  # type: ignore # pylint: disable=unreachable

    @classmethod
    async def all(cls) -> List[AsyncFaunaModel]:
        """


        Finds all documents of the model in FaunaDB.



        Returns:
        --------


        List[AsyncFaunaModel]:


            A list of instances of the model if found, None otherwise.
        """

        try:
            _q = cls.q()

            refs = (await _q(q.paginate(q.match(q.index(cls.__name__.lower())))))[
                "data"
            ]

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:
            logging.error(exc)

            return []

    @classmethod
    async def delete_unique(cls, field: str, value: Any) -> bool:
        """


                Deletes a document in FaunaDB by a unique field.



                Parameters:
                -----------
                field: str


                    The name of the unique field to use in the search.


                value: Any


                    The value of the unique field to use in the search.



                Returns:
                --------


                bool:
        t

                    True if the document is deleted, False otherwise.
        """

        try:
            _q = cls.q()

            ref = await _q(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )

            await _q(q.delete(ref))

            return True

        except AioFaunaException as exc:
            logging.error(exc)

            return False

    @classmethod
    async def delete(cls, ref: str) -> bool:
        """Delete a document by id"""

        try:
            await cls.q()(q.delete(q.ref(q.collection(cls.__name__.lower()), ref)))

            return True

        except AioFaunaException as exc:
            logging.error(exc)

            return False

    async def create(self) -> AsyncFaunaModel:
        """


        Creates a new document in FaunaDB.



        Parameters:
        -----------


        **kwargs:


            The data to create the new document with.



        Returns:
        --------


        Union[AsyncFaunaModel,BaseModel]:


            An instance of the model if created successfully, None otherwise.
        """

        try:
            for field in self.__fields__.values():
                if field.field_info.extra.get("unique"):
                    instance = await self.find_unique(
                        field.name, getattr(self, field.name)
                    )
                    if instance is None:
                        continue
                    await instance.__class__.q()(q.create(q.collection(self.__class__.__name__.lower()), {"data": self.dict()}))  # type: ignore
                    return instance  # type: ignore
            data = await self.__class__.q()(
                q.create(
                    q.collection(self.__class__.__name__.lower()), {"data": self.dict()}
                )
            )
            self.ref = data["ref"]["@ref"]["id"]
            self.ts = data["ts"] / 1000
            return self

        except AioFaunaException as exc:
            logging.error(exc)

            return None  # type: ignore # pylint: disable=unreachable

    @classmethod
    async def query(cls, query: str) -> List[AsyncFaunaModel]:
        """


        Queries FaunaDB using a string query and returns a list of instances of the model.



        Args:


            query (str): The query to use.



        Returns:


            List[AsyncFaunaModel]: A list of instances of the model if found, None otherwises.
        """

        try:
            refs = (await cls.q()(q.paginate(q.match(q.query(query)))))["data"]

            data = await asyncio.gather(*[cls.q()(q.get(ref)) for ref in refs])

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except AioFaunaException as exc:
            logging.error(exc)

            return []

    @classmethod
    async def update(cls, ref: str, **kwargs) -> AsyncFaunaModel:
        try:
            instance = await cls.find(ref)
            if isinstance(instance, cls):
                instance_updated = await cls.q()(
                    q.update(
                        q.ref(q.collection(cls.__name__.lower()), ref), {"data": kwargs}
                    )
                )
                return cls(
                    **{
                        **instance_updated["data"],
                        "ref": instance_updated["ref"]["@ref"]["id"],
                        "ts": instance_updated["ts"] / 1000,
                    }
                )
            else:
                raise ValueError(f"Field {ref} not found")
        except AioFaunaException as exc:
            logging.error(exc)
            raise ValueError(f"Field {ref} not found")

    async def save(self) -> AsyncFaunaModel:
        """


        Saves a document in FaunaDB.



        Returns:
        --------


        Union[AsyncFaunaModel,BaseModel]:


            An instance of the model if saved successfully, None otherwise.
        """

        if isinstance(self.ref, str) and len(self.ref) == 18:
            return await self.update(self.ref, kwargs=self.dict())
        return await self.create()


    @classmethod
    def gen_ts(cls):
        '''Health check endpoint'''
        schema = cls.schema()
        ts_config = {}
        for name_, field in schema["properties"].items():
            if field.get("type") in ["string", "number", "boolean", "integer", "null"]:    
                ts_config[name_] = field.get("type") if name_ in schema["required"] else f"{field['type']} | undefined"
            elif field.get("type") == "array":
                if field.get("items").get("type") in ["string", "number", "boolean", "integer", "null"]:
                    ts_config[name_] = f"{field.get('items').get('type')}[]" if name_ in schema["required"] else f"{field.get('items').get('type')}[] | undefined"
                elif field.get("items").get("$ref"):
                    ts_config[name_] = f"{field.get('items').get('$ref').split('/')[-1]}[]" if name_ in schema["required"] else f"{field.get('items').get('$ref').split('/')[-1]}[] | undefined"
                elif field.get("items").get("object"):
                    ts_config[name_] = f"{field.get('items').get('title')}[]" if name_ in schema["required"] else f"{field.get('items').get('title')}[] | undefined"
            elif field.get("type") == "object":
                ts_config[name_] = field.get("title") if name_ in schema["required"] else f"{field.get('title')} | undefined"
            elif field.get("$ref"):
                ts_config[name_] = field.get("$ref").split('/')[-1] if name_ in schema["required"] else f"{field.get('$ref').split('/')[-1]} | undefined"
            elif field.get("object"):   
                ts_config[name_] = field.get("title") if name_ in schema["required"] else f"{field.get('title')} | undefined"
            elif field.get("oneOf"):
                ts_config[name_] = " | ".join([f"{x.get('$ref').split('/')[-1]}" if x.get("$ref") else f"{x.get('title')}" for x in field.get("oneOf")])
            elif field.get("anyOf"):
                ts_config[name_] = " | ".join([f"{x.get('$ref').split('/')[-1]}" if x.get("$ref") else f"{x.get('title')}" for x in field.get("anyOf")])
            elif field.get("allOf"):
                ts_config[name_] = " & ".join([f"{x.get('$ref').split('/')[-1]}" if x.get("$ref") else f"{x.get('title')}" for x in field.get("allOf")])
            elif field.get("not"):
                ts_config[name_] = f"!{field.get('not').get('$ref').split('/')[-1]}" if field.get("not").get("$ref") else f"!{field.get('not').get('title')}"
            elif field.get("enum"):
                ts_config[name_] = " | ".join([f"{x}" for x in field.get("enum")])
            else:
                ts_config[name_] = "any"
                
        ts_type = f"type {cls.__name__} = {ts_config}"
        
        return ts_type.replace("'", '')
    

    @classmethod
    def gen_store(cls):
        return f"""
        import {cls.__name__} from '~/types/{cls.__name__.lower()}'
        import {{ defineStore, acceptHMRUpdate }} from 'pinia'
        
        export const {cls.__name__}Store = defineStore('{cls.__name__.lower()}', () => {{
            const state = reactive({{
                {cls.__name__.lower()}s: [] as {cls.__name__}[],
                {cls.__name__.lower()}: null as {cls.__name__} | null,
                loading: false,
                error: null as string | null,
                notifications: [] as {{message: string, status: 'success' | 'error' | 'warning' | 'info'}}[]
            }})
            
            const api = {{
                async fetch{cls.__name__}s() {{  
                    state.loading = true
                    try {{
                        const {{ data }} =  await useFetch("/api/{cls.__name__.lower()}s").json()
                        state.{cls.__name__.lower()}s = unref(data)
                    }} catch (error) {{
                        state.error = error.message
                    }} finally {{
                        state.loading = false
                    }}
                }},
                async fetch{cls.__name__}(id: string) {{
                    state.loading = true
                    try {{
                        const {{ data }} =  await useFetch("/api/{cls.__name__.lower()}s/" + id).json()
                        state.{cls.__name__.lower()} = unref(data)
                    }} catch (error) {{
                        state.error = error.message
                    }} finally {{
                        state.loading = false
                    }}
                }},
                async create{cls.__name__}({cls.__name__.lower()}: {cls.__name__}) {{
                    state.loading = true
                    try {{
                        const {{ data }} =  await useFetch("/api/{cls.__name__.lower()}s", {{
                            method: "POST",
                            headers: {{
                                "Content-Type": "application/json"
                            }},
                            body: JSON.stringify({cls.__name__.lower()})
                        }}).json()
                        state.{cls.__name__.lower()} = unref(data)
                    }} catch (error) {{
                        state.error = error.message
                    }} finally {{
                        state.loading = false
                    }}
                }},
                async update{cls.__name__}({cls.__name__.lower()}: {cls.__name__}) {{
                    state.loading = true
                    try {{
                        const {{ data }} =  await useFetch("/api/{cls.__name__.lower()}s/" + {cls.__name__.lower()}.ref, {{
                            method: "PUT",
                            headers: {{
                                "Content-Type": "application/json"
                            }},
                            body: JSON.stringify({cls.__name__.lower()})
                        }}).json()
                        state.{cls.__name__.lower()} = unref(data)
                    }} catch (error) {{
                        state.error = error.message
                    }} finally {{
                        state.loading = false
                    }}
                }},
                async delete{cls.__name__}({cls.__name__.lower()}: {cls.__name__}) {{
                    state.loading = true
                    try {{
                        const {{ data }} =  await useFetch("/api/{cls.__name__.lower()}s/" + {cls.__name__.lower()}.ref, {{
                            method: "DELETE"
                        }}).json()
                        state.{cls.__name__.lower()} = unref(data)
                    }} catch (error) {{
                        state.error = error.message
                    }} finally {{
                        state.loading = false
                    }}
                }}
            }}

        
            return {{
                state,
                api
            }}
        }})
        
        if (import.meta.hot) {{
            import.meta.hot.accept(acceptHMRUpdate({cls.__name__}Store, import.meta.hot))
        }}
        """
            
            


"""ASGI Middleware"""
import typing
import io
from aiohttp.web import Request
from aiofauna.api import Api

Scope = typing.MutableMapping[str, typing.Any]
Message = typing.MutableMapping[str, typing.Any]

Receive = typing.Callable[[], typing.Awaitable[Message]]
Send = typing.Callable[[Message], typing.Awaitable[None]]

ASGIApp = typing.Callable[[Scope, Receive, Send], typing.Awaitable[None]]


def aioasgi(app: Api) -> ASGIApp:
    async def asgi(scope: Scope, receive: Receive, send: Send) -> None:
        nonlocal app
        if scope["type"] != "http":
            raise RuntimeError("Only HTTP is supported")

        method = scope["method"]
        path = scope["path"]
        query_string = scope.get("query_string", b"").decode()
        headers = [[key.decode(), value.decode()] for key, value in scope["headers"]]
        body = b""

        content_length = int(scope.get("content_length", 0))

        while content_length:
            message = await receive()
            _body = message.get("body", b"")
            body += _body
            content_length -= len(_body)

        request = Request(
            method=method,
            path=path,
            query_string=query_string,
            headers=headers,
            body=body,
        )

        response = await app._handle(request)  # pylint: disable=protected-access

        body_length = response.body_length

        response_body = b""

        while body_length:
            if response.chunked:
                writer = response._payload_writer  # pylint: disable=protected-access
                if writer is None:
                    response_body = (
                        response._content_dict
                    )  # pylint: disable=protected-access
                    break
                buffer_size = writer.buffer_size
                with io.BytesIO() as buffer:
                    while buffer_size:
                        await writer.drain()
                        chunk = response._body  # pylint: disable=protected-access
                        if not chunk:  # pylint: disable=protected-access
                            break
                        if isinstance(chunk, bytes):
                            buffer.write(chunk)
                            buffer_size -= len(chunk)
                        elif isinstance(chunk, str):
                            buffer.write(chunk.encode())
                            buffer_size -= len(chunk.encode())
                        else:
                            raise RuntimeError("Invalid chunk type")
                    response_body += buffer.getvalue()
                    body_length -= len(response_body)
            else:
                response_body = (
                    response._content_dict
                )  # pylint: disable=protected-access
                break

        await send(
            {
                "type": "http.response.start",
                "status": response.status,
                "headers": response.headers,
            }
        )

        await send(
            {
                "type": "http.response.body",
                "body": response_body,
            }
        )

    return asgi



import io

import os

from typing import Optional as O, Dict as D, Any as A, AsyncGenerator as AG

from aiohttp import ClientSession

from dotenv import load_dotenv

from .errors import AioFaunaException

from .objects import Expr


from .json import FaunaJSONEncoder

load_dotenv()


def to_json(obj: Expr) -> str:
    """


    `to_json`


    Summary:


        Serialize a FaunaDB query.


    Args:


        obj: A FaunaDB query.


    Returns:


        A JSON string representing the FaunaDB query.
    """

    return FaunaJSONEncoder().encode(obj)


class AsyncFaunaClient(object):
    """


    `AsyncFaunaClient`


    Summary:


        Introducing support for coroutines and async/await syntax for FaunaDB python driver.


    Description:


        Using `_Expr` data type as an input for `AsyncFaunaClient.query` method.


        We can build fauna queries using `q` module methods and pass them to `AsyncFaunaClient.query` method


        as an instance of `_Expr` data type to execute them.


        The `_Expr` object will be serialized with `FaunaJSONEncoder` from json module so the FaunaDB API can understand it.


        The response will be deserialized by the `FaunaModel` class from orm module.
    """

    def __init__(self, secret=None):
        if secret is None:
            secret = os.getenv("FAUNA_SECRET")

        self.secret = secret

    async def query(self, expr: Expr) -> A:
        """


        `AsyncFaunaClient.query`


        Summary:


            Execute a FaunaDB query.



        Args:


            expr: A FaunaDB query.



        Returns:


            A FaunaDB response.
        """

        async with ClientSession() as session:
            async with session.post(
                "https://db.fauna.com",
                data=to_json(expr),
                headers={
                    "Authorization": f"Bearer {self.secret}",
                    "Content-type": "application/json",
                    "Accept": "application/json",
                },
            ) as response:
                try:
                    data = await response.json()

                    return data["resource"]

                except AioFaunaException as exc:
                    print(exc)

                    return None

    async def stream(self, expr: Expr) -> AG[A, None]:
        """

        `AsyncFaunaClient.stream`


        Summary:


                Streams a FaunaDB document.


        Args:


                expr: A FaunaDB query.


        Returns:


                An Event Stream Generator.
        """

        async with ClientSession() as session:
            async with session.post(
                "https://db.fauna.com",
                data=to_json(expr),
                headers={
                    "Authorization": f"Bearer {self.secret}",
                    "Content-type": "application/json",
                    "Accept": "text/event-stream",
                    "Keep-Alive": "timeout=5, max=900",
                    "Connection": "keep-alive",
                    "Cache-Control": "no-cache",
                    "X-Last-Seen-Txn": "0",
                    "X-Request-By": "aiofauna",
                    "X-Query-By": "aiofauna",
                },
            ) as response:
                with io.StringIO() as buffer:
                    async for chunk in response.content.iter_chunked(1024):
                        buffer.write(chunk.decode())

                        buffer.seek(0)

                        for line in buffer:
                            if line == "\r\n":
                                yield buffer.getvalue()

                                buffer.seek(0)

                                buffer.truncate()

                            else:
                                yield line


class HTTPClient:
    """

    Generic HTTP Client

    """

    async def fetch(
        self,
        url: str,
        method: str = "GET",
        headers: O[D[str, str]] = None,
        data: O[D[str, A]] = None,
    ) -> A:
        """
        Generic function to retrieve data from an URL in json format
        """
        if method in ["GET", "DELETE"]:
            async with ClientSession() as session:
                async with session.request(method, url, headers=headers) as response:
                    return await response.json()
        elif method in ["POST", "PUT", "PATCH"]:
            async with ClientSession() as session:
                async with session.request(
                    method, url, headers=headers, json=data
                ) as response:
                    return await response.json()
        else:
            raise ValueError("Invalid method")

    async def text(
        self,
        url: str,
        method: str = "GET",
        headers: O[D[str, str]] = None,
        data: O[D[str, A]] = None,
    ) -> str:
        """
        Generic function to retrieve data from an URL in text format
        """
        if method in ["GET", "DELETE"]:
            async with ClientSession() as session:
                async with session.request(method, url, headers=headers) as response:
                    return await response.text()
        elif method in ["POST", "PUT", "PATCH"]:
            async with ClientSession() as session:
                async with session.request(
                    method, url, headers=headers, json=data
                ) as response:
                    return await response.text()
        else:
            raise ValueError("Invalid method")

    async def blob(
        self,
        url: str,
        method: str = "GET",
        headers: O[D[str, str]] = None,
        data: O[D[str, A]] = None,
    ) -> bytes:
        """
        Generic function to retrieve data from an URL in binary format
        """
        if method in ["GET", "DELETE"]:
            async with ClientSession() as session:
                async with session.request(method, url, headers=headers) as response:
                    return await response.read()
        elif method in ["POST", "PUT", "PATCH"]:
            async with ClientSession() as session:
                async with session.request(
                    method, url, headers=headers, json=data
                ) as response:
                    return await response.read()
        else:
            raise ValueError("Invalid method")

    async def stream(
        self,
        url: str,
        method: str = "GET",
        headers: O[D[str, str]] = None,
        data: O[D[str, A]] = None,
    ) -> AG[str, None]:
        """
        Generic function to retrieve data from an URL in streaming format
        """
        if method in ["GET", "DELETE"]:
            async with ClientSession() as session:
                async with session.request(method, url, headers=headers) as response:
                    async for chunk in response.content.iter_chunked(1024):
                        yield chunk.decode("utf-8")
        elif method in ["POST", "PUT", "PATCH"]:
            async with ClientSession() as session:
                async with session.request(
                    method, url, headers=headers, json=data
                ) as response:
                    async for chunk in response.content.iter_chunked(1024):
                        yield chunk.decode("utf-8")
        else:
            raise ValueError("Invalid method")



from __future__ import annotations
from typing import List
from pydantic import Field, BaseModel  # pylint: disable=no-name-in-module


class Graph(BaseModel):

    """
    `Graph` is a class that represents a graph data structure. It contains a list of nodes and a list of edges. It also contains methods for adding nodes and edges to the graph, as well as methods for checking if the graph is cyclic and performing a topological sort on the graph.
    """

    nodes: List[int] = Field(default_factory=list)
    edges: List[List[int]] = Field(default_factory=list)

    def add_node(self, node: int):
        """
        The add_node method adds a node to the graph.
        """
        self.nodes.append(node)

    def add_edge(self, edge: List[int]):
        """
        The add_edge method adds an edge to the graph.
        """
        self.edges.append(edge)

    @property
    def is_cyclic(self) -> bool:
        """
        The is_cyclic property checks if the graph is cyclic.
        """

        return self._is_cyclic(self.nodes, self.edges)

    @staticmethod
    def _is_cyclic(nodes: List[int], edges: List[List[int]]) -> bool:
        """
        The _is_cyclic method iterates through each node in the graph and checks if the graph is cyclic using the _is_cyclic_helper method.
        """
        visited = set()
        for node in nodes:
            if node not in visited:
                if Graph._is_cyclic_helper(node, visited, edges):
                    return True
        return False

    @staticmethod
    def _is_cyclic_helper(node: int, visited: set, edges: List[List[int]]) -> bool:
        """
        The _is_cyclic_helper method performs a depth-first search on the node and its neighbors. If a node has already been visited, then the graph is cyclic.
        """
        if node in visited:
            return True
        visited.add(node)
        for edge in edges:
            if node in edge:
                if Graph._is_cyclic_helper(edge[1], visited, edges):
                    return True
        return False

    def topological_sort(self) -> List[int]:
        """
        The topological_sort method performs a topological sort on the graph using a modified depth-first search algorithm. It initializes a set of visited nodes and a list of sorted nodes, and then iterates through each node in the graph.
        """
        return self._topological_sort(self.nodes, self.edges)

    @staticmethod
    def _topological_sort(nodes: List[int], edges: List[List[int]]) -> List[int]:
        """
        _topological_sort_helper method to perform depth-first search on the node and its neighbors. Once all nodes have been visited, the method returns the sorted nodes.
        """
        visited = set()
        sorted_nodes = []
        for node in nodes:
            if node not in visited:
                Graph._topological_sort_helper(node, visited, edges, sorted_nodes)
        return sorted_nodes

    @staticmethod
    def _topological_sort_helper(
        node: int, visited: set, edges: List[List[int]], sorted_nodes: List[int]
    ):
        """
        The _topological_sort_helper method performs depth-first search on a node and its neighbors recursively. It adds the node to the visited set and then recursively visits its neighbors. Once all neighbors have been visited, it appends the node to the sorted list.
        """
        visited.add(node)
        for edge in edges:
            if node in edge:
                Graph._topological_sort_helper(edge[1], visited, edges, sorted_nodes)
        sorted_nodes.append(node)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiofauna']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'Markdown>=3.4.3,<4.0.0',
 'Pygments>=2.15.1,<3.0.0',
 'aiohttp-devtools>=1.0.post0,<2.0',
 'aiohttp-sse>=2.1.0,<3.0.0',
 'aiohttp>=3.8.4,<4.0.0',
 'aiohttp_cors>=0.7.0,<0.8.0',
 'iso8601>=1.1.0,<2.0.0',
 'pydantic[dotenv]>=1.10.7,<2.0.0']

setup_kwargs = {
    'name': 'aiofauna',
    'version': '0.1.9',
    'description': '',
    'long_description': '---\ntitle: AioFauna\n---\n# AioFauna\n\n🚀 Introducing aiofauna: A full-stack framework built on top of Aiohttp, Pydantic and FaunaDB.\n\n🔥 Inspired by FastAPI focuses on Developer Experience, Productivity and Versatility.\n\n🌟 Features:\n\n✅ Supports Python 3.7+, comes with an opinionated ODM (Object Document Mapper) out of the box for FaunaDB that abstracts out complex FQL expressions into pythonic, fully typed asynchronous methods for all CRUD operations.\n\n✅ Performant and scalable: Built on top of Aiohttp a powerful http server and client and FaunaDB an scalable serverless database for modern applications.\n\n✅ Async/await coroutines: Leverage the power of async programming for enhanced performance and responsiveness, being ASGI compliant is compatible with most async python frameworks.\n\n✅ Automatic Swagger UI generation: Automatic generation of interactive Swagger UI documentation for instant testing of your API.\n\n✅ SSE (Server Sent Events): Built-in support for SSE (Server Sent Events) for real-time streaming of data from FaunaDB to your application.\n\n✅ Robust data validation: Built-in support for Pydantic models for robust data validation and serialization.\n\n✅ Auto-provisioning: Automatic management of indexes, unique indexes, and collections with FaunaModel ODM.\n\n✅ Full JSON communication: Custom encoder to ensure seamless data exchange between your application and FaunaDB backend.\n\n✅ Markdown and Jinja support with live reload: experiment an smooth frontend devserver experience without leaving your backend code.\n\n✅ Inspired by fastapi, you will work with almost the same syntax and features like path operations, path parameters, query parameters, request body, status codes and more.\n\n💡 With aiofauna, you can build fast, scalable, and reliable modern applications, while building seamless integrations thanks to the fastest http client aiohttp and the most versatile database FaunaDB, you will enjoy integrating with third party services such as APIs, Data Sources and Cloud Services.\n\n📚 Check out the aiofauna library, and start building your next-gen applications today! 🚀\n\n#Python #FaunaDB #Async #Pydantic #aiofauna\n\n⚙️ If you are using a synchronous framework check out [Faudantic](https://github.com/obahamonde/faudantic) for a similar experience with FaunaDB and Pydantic.\n\n📚 [Documentation](https://obahamonde-aiofauna-docs.smartpro.solutions) (Built with aiofauna)\n\n📦 [PyPi](https://pypi.org/project/aiofauna/)\n\n📦 [GitHub](https://github.com/obahamonde/aiofauna)\n\n📦 [Demo](https://aiofaunastreams-fwuw7gz7oq-uc.a.run.app/) (Real time Latency Monitoring between FaunaDB and Google Cloud Run)\n\n\n',
    'author': 'Oscar Bahamonde',
    'author_email': 'oscar.bahamonde.dev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/obahamonde/aiofauna',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
